/// Formats that the default glycin loaders support
pub const DEFAULT_MIME_TYPES: &[&str] = &[
    // image-rs
    "image/jpeg",
    "image/png",
    "image/gif",
    "image/webp",
    "image/tiff",
    "image/x-tga",
    "image/vnd-ms.dds",
    "image/x-dds",
    "image/bmp",
    "image/vnd.microsoft.icon",
    "image/vnd.radiance",
    "image/x-exr",
    "image/x-portable-bitmap",
    "image/x-portable-graymap",
    "image/x-portable-pixmap",
    "image/x-portable-anymap",
    "image/x-qoi",
    // HEIF
    "image/avif",
    "image/heif",
    // JXL
    "image/jxl",
    // SVG
    "image/svg+xml",
    "image/svg+xml-compressed",
];
